import os
import shutil
import re
import datetime
import git
import configparser
import time
import subprocess
import tempfile
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.firefox.service import Service
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
import pyautogui

def copy_folder(src, dst):
    if os.path.exists(dst):
        shutil.copytree(src, dst,dirs_exist_ok=True)
    else:
        shutil.copytree(src, dst)

def remove_folder(dst):
    if os.path.exists(dst):
        shutil.rmtree(dst)

def is_exceptional_md(filename):
    exceptions = ['Markdown-Picture-Resize-and-Location.md']
    return filename in exceptions

def process_check(filepath,log):

    with open(filepath, 'r', encoding='utf-8') as file:
        lines = file.readlines()

    for i, line in enumerate(lines):
        matches = re.findall(r'[^!]\+?\[[^[#]*?\]\(.*?\)', line)
        for match in matches:
            if log==1:
                print("standard-link line=" + str(i) + " path=" + filepath + " match=" + match + " conetent=" + line.strip()[0:30] + "...")

    for i, line in enumerate(lines):
        matches = re.findall(r'!\[[^[#]*?\]\(.*?\)', line)
        for match in matches:
            if log==1:
                print("standard-link-display line=" + str(i) + " path=" + filepath + " match=" + match + " conetent=" + line.strip()[0:30] + "...")

    for i, line in enumerate(lines):
        matches = re.findall(r'!\[\[([^\[]*?#[^\[]*?)\]\]', line)
        for match in matches:
            if log==1:
                print("ob-link-display line=" + str(i) + " path=" + filepath + " match=" + match + " conetent=" + line.strip()[0:30] + "...")

    for i, line in enumerate(lines):
        matches = re.findall(r'\\begin{matrix\*}', line)
        for match in matches:
            if log==1:
                print("katex-matrix* line=" + str(i) + " path=" + filepath + " match=" + match + " conetent=" + line.strip()[0:30] + "...")

    for i, line in enumerate(lines):
        matches = re.findall(r'\\begin{align\*}', line)
        for match in matches:
            if log==1:
                print("katex-align* line=" + str(i) + " path=" + filepath + " match=" + match + " conetent=" + line.strip()[0:30] + "...")

def get_file_properties_time(file_path):
    
    # 获取文件创建时间
    create_time = os.path.getctime(file_path)
    # 获取文件修改时间
    modify_time = os.path.getmtime(file_path)

    # 将时间戳转换为可读格式
    create_time = datetime.datetime.fromtimestamp(create_time)
    modify_time = datetime.datetime.fromtimestamp(modify_time)

    # 将时间戳转换为wiki.js格式
    create_time=create_time.strftime('%Y-%m-%dT%H:%M:%S.%f')
    create_time=re.sub(r'\d{3}$',r'+8:00',create_time)
    modify_time=modify_time.strftime('%Y-%m-%dT%H:%M:%S.%f')
    modify_time=re.sub(r'\d{3}$',r'+8:00',modify_time)

    return create_time, modify_time

def process_properties(filepath):
    with open(filepath, 'r', encoding='utf-8') as file:
        lines = file.readlines()

    prop_start = None
    prop_end = None

    for i, line in enumerate(lines):
        if line.strip() == '---':
            if prop_start is None:
                prop_start = i
            else:
                prop_end = i
                break

    if prop_start is not None and prop_end is not None:
        del lines[prop_start:prop_end + 1]

    filename = os.path.basename(filepath)
    titleValue = filename[:-3]  # Removing '.md' extension
    #dateValue = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + '+8:00'  # ISO 8601 format
    #dateCreatedValue = dateValue
    dateCreatedValue, dateValue = get_file_properties_time(filepath)

    new_properties = f"---\n" \
                     f"title: {titleValue}\n" \
                     f"description:\n" \
                     f"published: true\n" \
                     f"date: {dateValue}\n" \
                     f"tags:\n" \
                     f"editor: Ring\n" \
                     f"dateCreated: {dateCreatedValue}\n" \
                     f"---\n"

    lines.insert(0, new_properties)

    with open(filepath, 'w', encoding='utf-8') as file:
        file.writelines(lines)

def process_attachments(filepath,search_path,log):
    with open(filepath, 'r', encoding='utf-8') as file:
        content = file.read()
        
    matches = re.findall(r'!\[\[([^#]*?)\]\]', content)

    for match in matches:
        if len(match.split('|')) > 1:
            attachment_name = match.split('|')[0]
            attachment_size = match.split('|')[1]
            if len(attachment_size.split('x'))==1:
                attachment_size = " ="+attachment_size+"x"
        else:
            attachment_name = match.split('|')[0]
            attachment_size = ""

        for root, dirs, files in os.walk(search_path):
            if attachment_name in files:
                if log==1:
                    print("attachments "+filepath+" "+attachment_name)
                attachment_path = os.path.relpath(os.path.join(root, attachment_name), start=search_path)
                new_path = "/" + attachment_path.replace('\\', '/')
                content = content.replace(f'![[{match}]]', f'![{attachment_name}]({new_path}{attachment_size})')
    
    with open(filepath, 'w', encoding='utf-8') as file:
        file.write(content)


def process_inlink(filepath,search_path,log):

    with open(filepath, 'r', encoding='utf-8') as file:
        content = file.read()

    matches = re.findall(r'[^!]\+?\[\[([^\[,]*?)\]\]', content)

    for match in matches:
        if len(match.split('#')) > 1:
            link_name = match.split('#')[0]
            link_head = match.split('#')[1]
            link_head = "#"+link_head
            link_head_use=link_head.replace(" ", "-").replace(".", "").lower()
            link_head_use=re.sub(r'#(\d+)', r'#h-\1', link_head_use)
            cleaned_string = re.findall(r'[\u4e00-\u9fa5a-zA-Z0-9#-]+', link_head_use)
            link_head_use=''.join(cleaned_string)
            
        else:
            link_name = match.split('#')[0]
            link_head = ""
            link_head_use=""

        if len(link_name)==0 and len(link_head)>1:
            if log==1:
                print("inlink-head "+filepath+" "+link_name+" "+link_head_use)
            content = content.replace(f'[[{match}]]', f'[{link_head}]({link_head_use})')
            
        for root, dirs, files in os.walk(search_path):
            if link_name in files:
                if log==1:
                    print("inlink-attach "+filepath+" "+link_name+" "+link_head)
                link_path = os.path.relpath(os.path.join(root, link_name), start=search_path)
                new_path = "/" + link_path.replace('\\', '/')
                content = content.replace(f'[[{match}]]', f'[{link_name}]({new_path})')
                
            link_md_name = link_name + ".md"
            if link_md_name in files:
                if log==1:
                    print("inlink-md-head "+filepath+" "+link_md_name+link_head_use)
                link_md_name = link_name.replace(" ","-").replace(".","") + ".md"
                link_path = os.path.relpath(os.path.join(root, link_md_name), start=search_path)
                new_path = "/" + link_path.replace('\\', '/')
                new_path = re.sub(r'\.md$', '', new_path)
                content = content.replace(f'[[{match}]]', f'[{link_name}{link_head}]({new_path}{link_head_use})')

    with open(filepath, 'w', encoding='utf-8') as file:
        file.write(content)


def process_head_rename(filepath,log):
    
    with open(filepath, 'r', encoding='utf-8') as file:
        content = file.read()

    matches = re.findall(r'^(#+\s+.*)$', content, re.MULTILINE)
    
    for match in matches:
        total = len(match.split(" "))
        for num in range(0,total-1):
            if num==0:
                new_head=match.split(" ")[num] + " " + match.split(" ")[num+1]
            else:
                new_head=new_head + "-" +match.split(" ")[num+1]
        if log==1:
            print("head " + filepath + " " + match + " -> " + new_head)  
        content = re.sub(f'{match}', f'{new_head}', content, flags=re.MULTILINE)
        print("finish")
          

    with open(filepath, 'w', encoding='utf-8') as file:
        file.write(content)

def process_latex(filepath,log):
    
    with open(filepath, 'r', encoding='utf-8') as file:
        content = file.read()

    matches = re.findall(r'\\begin{align}', content)
    for match in matches:
        if log==1:
            print("latex "+filepath+" "+match)
        content = content.replace(r'\begin{align}', r'\begin{aligned}')
        content = content.replace(r'\end{align}', r'\end{aligned}')


    matches = re.findall(r'\\begin{gather}', content)
    for match in matches:
        if log==1:
            print("latex "+filepath+" "+match)
        content = content.replace(r'\begin{gather}', r'\begin{gathered}')
        content = content.replace(r'\end{gather}', r'\end{gathered}')
    
    while re.findall(r'{{', content):
        if log==1:
            print("latex "+filepath+" {{ ")
        content = re.sub(r'{{', r'{ {', content)
        
    while re.findall(r'}}', content):
        if log==1:
            print("latex "+filepath+" }} ")
        content = re.sub(r'}}', r'} }', content)
            
    with open(filepath, 'w', encoding='utf-8') as file:
        file.write(content)


def process_github(source_path,repo_path):

    # 配置 repo
    repo = git.Repo(repo_path)
    config = configparser.ConfigParser()
    config.read(repo_path + '/.git/config')
    origin = repo.remote(name='wiki')

    # 同步 pull remote repo
    origin.pull('main')

    # 完全更新 local repo
    for item in os.listdir(repo_path):
        if item != '.git':
            item_path = os.path.join(repo_path, item)
            if os.path.isfile(item_path):
                os.remove(item_path)
            else:
                shutil.rmtree(item_path)
    copy_folder(source_path,repo_path)

    # 提交 commit 新增/删除
    repo.git.add('.')
    msg = datetime.datetime.now().strftime('%Y-%m-%d %H:%M')
    repo.git.commit('-a', '-m', msg)

    # 推送 git pull origin main
    origin.push('main')

def generate_markdown(folder_path):
    folder_name = os.path.basename(folder_path)
    markdown_file = os.path.join(os.path.dirname(folder_path), f"{folder_name}.md")
    
    with open(markdown_file, "w", encoding="utf-8") as f:
        f.write(f"# {folder_name}\n\n")
        
        for item in os.listdir(folder_path):
            item_path = os.path.join(folder_path, item)
            if os.path.isdir(item_path):
                f.write(f"- [[{item}]]\n")
            elif os.path.isfile(item_path):
                filename, _ = os.path.splitext(item)
                f.write(f"- [[{filename}]]\n")


def process_deleteAllPages(db_host, db_port, db_name, db_user, db_password):

    sql_commands = """
    TRUNCATE TABLE "pages" CASCADE;
    TRUNCATE TABLE "pageTree" CASCADE;
    TRUNCATE TABLE "assetData" CASCADE;
    TRUNCATE TABLE "assetFolders" CASCADE;
    TRUNCATE TABLE "assets" CASCADE;
    TRUNCATE TABLE "pageHistory" CASCADE;
    TRUNCATE TABLE "pageLinks" CASCADE;
    ALTER SEQUENCE "pages_id_seq" RESTART WITH 1;
    ALTER SEQUENCE "assetFolders_id_seq" RESTART WITH 1;
    ALTER SEQUENCE "assets_id_seq" RESTART WITH 1;
    ALTER SEQUENCE "pageHistory_id_seq" RESTART WITH 1;
    ALTER SEQUENCE "pageLinks_id_seq" RESTART WITH 1;
    """

    with tempfile.NamedTemporaryFile(delete=False, mode='w', encoding='utf-8', suffix='.sql') as temp_sql_file:
        temp_sql_file.write(sql_commands)
        temp_sql_file_path = temp_sql_file.name

    try:
        env = os.environ.copy()
        env['PGPASSWORD'] = db_password
        env['PGCLIENTENCODING'] = 'UTF8'  # 设置 PostgreSQL 输出编码

        result = subprocess.run(
            ["psql", "-h", db_host, "-p", db_port, "-d", db_name, "-U", db_user, "-f", temp_sql_file_path],
            env=env,
            capture_output=True
        )

        try:
            stdout = result.stdout.decode('utf-8')
            stderr = result.stderr.decode('utf-8')
        except UnicodeDecodeError:
            # 如果出现解码错误，尝试使用其他编码解码
            stdout = result.stdout.decode('latin1')
            stderr = result.stderr.decode('latin1')

        if result.returncode != 0:
            raise Exception(f"SQL脚本执行失败，退出代码：{result.returncode}")
        else:
            print("SQL脚本执行成功。")

    except Exception as e:
        print(f"发生错误：{str(e)}")
        print("详细错误信息：", stderr)

    finally:
        if os.path.exists(temp_sql_file_path):
            os.remove(temp_sql_file_path)
    
        env.pop('PGPASSWORD', None)
        print("finish process_deleteAllPages")


def process_import(username, password, site='http://127.0.0.1:5960/login', gecko_path='D:\\geckodriver.exe'):
    # 设置 WebDriver
    service = Service(gecko_path)
    driver = webdriver.Firefox(service=service)

    try:
        # 打开 Wiki.js 登录页面
        driver.get(site)

        # 等待用户名输入框出现
        username_input = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.ID, 'input-20'))  # 替换为实际的选择器
        )

        # 等待密码输入框出现
        password_input = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.ID, 'input-22'))  # 替换为实际的选择器
        )

        # 输入用户名和密码并提交
        username_input.send_keys(username)
        password_input.send_keys(password)
        password_input.send_keys(Keys.RETURN)

        # 点击管理按钮
        adminstration_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, 'i.v-icon.mdi.mdi-cog'))
        )
        adminstration_button.click()


        # 点击 Storage 按钮
        storage_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, '//div[text()="Storage"]'))
        )
        storage_button.click()
        time.sleep(1)
        
        # 点击 Local storage 按钮
        local_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, '//div[text()="Local storage on disk or network shares."]'))
        )
        local_button.click()
        time.sleep(1)

        # 点击 Import Everything 按钮
        import_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, '/html/body/div/div/div/main/div/div/div/div[3]/div/div[2]/form/div[11]/div/div[3]/div/div/button/span'))
        )
        import_button.click()
        time.sleep(1)

        action_completed_element = WebDriverWait(driver, 4800).until(
            EC.presence_of_element_located((By.XPATH, "//span[text()='Action completed.']"))
        )

        # 点击 Utilities 按钮
        utilities_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, '//div[text()="Utilities"]'))
        )
        utilities_button.click()
        time.sleep(1)
        
        # 点击 Content 按钮
        content_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, '//div[text()="Content"]'))
        )
        content_button.click()
        time.sleep(1)

        # 点击 rebuild Page Tree 按钮
        rebuild_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, '/html/body/div/div/div/main/div/div/div/div[3]/div/div/button[1]/span/span'))
        )
        rebuild_button.click()
        time.sleep(1)

        action_completed_element = WebDriverWait(driver, 4800).until(
            EC.presence_of_element_located((By.XPATH, "//span[text()='Page Tree rebuilt successfully.']"))
        )
        
        # 点击 Rerender All Pages 按钮
        rerender_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, '/html/body/div/div/div/main/div/div/div/div[3]/div/div/button[2]/span/span'))
        )
        rerender_button.click()
        time.sleep(1)

        action_completed_element = WebDriverWait(driver, 4800).until(
            EC.presence_of_element_located((By.XPATH, "//span[text()='All pages have been rendered successfully.']"))
        )
        
        # 点击 Search Engine 按钮
        storage_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, '//div[text()="Search Engine"]'))
        )
        storage_button.click()
        time.sleep(1)
        
        # 点击 REBUILD INDEX 按钮
        rerender_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, '/html/body/div/div/div/main/div/div/div/div[1]/div/button[2]/span/span'))
        )
        rerender_button.click()
        time.sleep(1)
        
        action_completed_element = WebDriverWait(driver, 4800).until(
            EC.presence_of_element_located((By.XPATH, "//span[text()='Index rebuilt successfully.']"))
        )
        

    finally:
        print("finish process_import")
        time.sleep(1)
        # 截取当前桌面的截图保存到指定路径
        # screenshot = pyautogui.screenshot()
        # screenshot.save("./finish.png")
        # 关闭浏览器的驱动程序和相关的资源，但不关闭浏览器窗口
        driver.quit()

def process_countMdFiles(dirPath):
    md_count = 0
    for root, dirs, files in os.walk(dirPath):
        for file in files:
            if file.endswith('.md'):
                md_count += 1
    print(f"Total {md_count} .md files in {dirPath}")

def process_quit():
    print("\nquit in 5s...")
    time.sleep(1)
    print("quit in 4s...")
    time.sleep(1)
    print("quit in 3s...")
    time.sleep(1)
    print("quit in 2s...")
    time.sleep(1)
    print("quit in 1s...")
    time.sleep(1)
    print("quit in 0s...")
    time.sleep(1)


def main():
    print("--------------------------------------")
    print(" Obsidian markdown files to Wiki.js")
    print(" Author: Ring_LU")
    print("--------------------------------------\n")

    ob_path = r'D:\DockerServer\NextCloudDrive\obsidian\analog-ic'
    tmp_path = "./tmp"
    repo_path = "./analog-ic-wiki"

    remove_folder(tmp_path)
    copy_folder(ob_path,tmp_path)
    remove_folder(tmp_path+'/.obsidian')
    remove_folder(tmp_path+'/.trash')
    remove_folder(tmp_path+'/白板无边记')

    for root, dirs, _ in os.walk(tmp_path, topdown=True):
        dirs[:] = [d for d in dirs if d != "pictures"]  # 排除名为"pictures"的文件夹
        for dir_name in dirs:
            generate_markdown(os.path.join(root, dir_name))                       
    print("finish generate_markdown...")

    
    for root, dirs, files in os.walk(tmp_path):
        for filename in files:
            filepath = os.path.join(root, filename)
            if filename.endswith('.md'):
                process_check(filepath,0)
                process_latex(filepath,0)
            if filepath.endswith(".md") and not is_exceptional_md(filename):
                process_properties(filepath)
                process_inlink(filepath,tmp_path,0)
                process_attachments(filepath,tmp_path,0)
    print("finish process_latex...")
    

    #process_github(tmp_path,repo_path)
    process_deleteAllPages(db_host="localhost", db_port = "5432", db_name = "****", db_user = "****", db_password="****")
    process_import('****', '****', 'http://127.0.0.1:5960/login',)
    process_countMdFiles(tmp_path)
    #remove_folder(tmp_path)
    process_quit()
    
if __name__ == "__main__":
    main()
